library(quantmod)
library(dplyr)
library(ggplot2)
library(forecast)
library(MASS)
library(tseries)
library(rugarch)
library(caret)
library(randomForest)
library(e1071)
library(xgboost)
library(parallel)
library(Metrics)
library(PerformanceAnalytics)
library(lubridate)
library(shiny)
library(plotly)
library(dplyr)
library(tidyr)
library(shinyjs)
library(keras)
process_data <- function(df) {
  # Technical indicators
  # Moving Averages (MA)
  df$sma_20 <- SMA(df$ClosingPrice, n = 20)
  df$sma_50 <- SMA(df$ClosingPrice, n = 50)
  
  # Relative Strength Index (RSI)
  df$rsi <- RSI(df$ClosingPrice)
  # Moving Average Convergence Divergence (MACD)
  df$macd <- as.data.frame(MACD(df$ClosingPrice))$macd
  
  df <- na.omit(df)
  
  # Derived features
  # Convert data frame to xts object with Date as index
  xts_data <- xts(df[, -1], order.by = as.Date(df$Date))
  
  # # Calculate daily returns
  df$returns <- as.data.frame(dailyReturn(xts_data))$daily.returns
  
  
  # Log Returns
  df$log_returns <- log(1 + df$returns)
  df <- na.omit(df)
  
  # Convert date to numeric for model input
  df$date_num <- as.numeric(as.POSIXct(df$Date))
  return(df)
}

downloadStockData <- function(stock_symbols, start_date, end_date) {
  # Create an empty data frame
  stock_data <-
    data.frame(Date = character(), ClosingPrice = numeric())
  
  # Download stock data using quantmod
  tryCatch({
    getSymbols(
      stock_symbols,
      from = start_date,
      to = end_date,
      periodicity = "daily"
    )
    
    # Extract closing prices for each stock symbol
    stock_data <- lapply(stock_symbols, function(symbol) {
      closing_price <- Cl(get(symbol))
      data.frame(Date = index(closing_price),
                 ClosingPrice = as.vector(closing_price))
    })
    
    # Merge the stock data into a single data frame
    stock_data <-
      Reduce(function(x, y)
        merge(x, y, by = "Date", all = TRUE), stock_data)
  },
  error = function(e) {
    showNotification("Stock data download failed. Please try again with a different stock.",
                     type = "warning")
  })
  
  return(stock_data)
}

ui <- fluidPage(
  useShinyjs(),
  # Enable shinyjs
  
  titlePanel("Stock Forecasting Dashboard"),
  sidebarLayout(
    sidebarPanel(
      dateRangeInput(
        "date_range",
        "Select Date Range:",
        start = "2010-01-01",
        end = "2023-12-31"
      ),
      textInput("stock_symbol", "Enter Stock Symbol:"),
      selectInput(
        "model_name",
        "Select Model for Predictions:",
        choices = c(
          "Linear Regression",
          "ARIMA",
          "GARCH",
          "Random Forest",
          "SVM",
          "Gradient Boosting",
          "Recurrent Neural Networks",
          "LSTM"
        )
      ),
      conditionalPanel(
        condition = "input.model_name == 'ARIMA'",
        numericInput("arima_order_1", "ARIMA Order (p):", value = 1),
        numericInput("arima_order_2", "ARIMA Order (d):", value = 1),
        numericInput("arima_order_3", "ARIMA Order (q):", value = 1),
        numericInput("arima_seasonal_order_1", "ARIMA Seasonal Order (P):", value = 0),
        numericInput("arima_seasonal_order_2", "ARIMA Seasonal Order (D):", value = 0),
        numericInput("arima_seasonal_order_3", "ARIMA Seasonal Order (Q):", value = 0),
        numericInput("arima_seasonal_period", "ARIMA Seasonal Period:", value = 12),
        numericInput("arima_lambda", "ARIMA Lambda:", value = 0.4)
      ),
      conditionalPanel(
        condition = "input.model_name == 'GARCH'",
        numericInput("garch_order_1", "GARCH Order (p):", value = 0),
        numericInput("garch_order_2", "GARCH Order (q):", value = 1),
        numericInput("garch_arma_order_1", "GARCH ARMA Order (p):", value = 1),
        numericInput("garch_arma_order_2", "GARCH ARMA Order (q):", value = 0)
      ),
      conditionalPanel(
        condition = "input.model_name == 'Random Forest'",
        numericInput("rf_mtry", "Random Forest mtry:", value = 2)
      ),
      conditionalPanel(
        condition = "input.model_name == 'SVM'",
        textInput("svm_c", "SVM C:", value = "1"),
        textInput("svm_sigma", "SVM Sigma:", value = "0.1")
      ),
      conditionalPanel(
        condition = "input.model_name == 'Gradient Boosting'",
        numericInput("gb_nrounds", "Gradient Boosting nrounds:", value = 1000),
        numericInput("gb_eta", "Gradient Boosting eta:", value = 0.01),
        numericInput("gb_lambda", "Gradient Boosting lambda:", value = 1),
        numericInput("gb_alpha", "Gradient Boosting alpha:", value = 0)
      ),
      conditionalPanel(condition = "input.model_name == 'Recurrent Neural Networks'",
                       numericInput("rnn_units", "RNN Units:", value = 50)),
      conditionalPanel(
        condition = "input.model_name == 'LSTM'",
        numericInput("lstm_units", "LSTM Units:", value = 50)
      ),
      actionButton("download_button", "Download Stock Data"),
      actionButton("fit_model_button", "Fit Model", disabled = TRUE),
      textOutput("fit_model_status")
    ),
    mainPanel(tabsetPanel(
      tabPanel("Plot", plotlyOutput("stock_plot")),
      tabPanel("Data", dataTableOutput("downloaded_data")),
      tabPanel("Model Prediction", plotlyOutput("model_prediction"))
      
    ))
  )
)

server <- shinyServer(function(input, output, session) {
  # Reactive value to track if data has been downloaded and model has been fit
  data_downloaded <- reactiveVal(FALSE)
  model_fitted <- reactiveVal(FALSE)
  
  # Define stock_data variable
  stock_data <- reactiveVal()
  
  # Reactive function to download stock data based on user input
  # Download and display sample data
  observeEvent(input$download_button, {
    stock_data_val <-
      downloadStockData(input$stock_symbol, input$date_range[1], input$date_range[2])
    if (!is.null(stock_data_val) && nrow(stock_data_val) > 0) {
      data_downloaded(TRUE)  # Set data_downloaded to TRUE
      model_fitted(FALSE)   # Reset model_fitted to FALSE
      
      stock_data(stock_data_val)  # Update stock_data
      
      stock_name <- toupper(input$stock_symbol)
      output$stock_plot <- renderPlotly({
        req(stock_data())  # Ensure stock_data is available
        stock_data() %>%
          gather(Symbol, ClosingPrice, -Date) %>%
          plot_ly(
            x = ~ Date,
            y = ~ ClosingPrice,
            color = ~ Symbol,
            type = "scatter",
            mode = "lines"
          ) %>%
          layout(
            title = paste("Stock Data -", stock_name),
            xaxis = list(
              title = "Date",
              tickfont = list(weight = "bold")
            ),
            yaxis = list(
              title = "Closing Price",
              tickfont = list(weight = "bold")
            )
          )
      })
      output$downloaded_data <- renderDataTable({
        req(stock_data())  # Ensure stock_data is available
        stock_data()
      })
      # Enable the Fit Model button
      shinyjs::enable("fit_model_button")
    } else {
      showNotification("Please download stock data first.", type = "warning")
    }
  })
  
  # Fit Model button action
  observeEvent(input$fit_model_button, {
    # Disable the Fit Model button
    shinyjs::disable("fit_model_button")
    # Show "Model Fitting Running" text
    output$fit_model_status <-
      renderText("Model Fitting Running...")
    req(stock_data())  # Ensure stock_data is available
    if (!is.null(stock_data()) && nrow(stock_data()) > 0) {
      processed_data <-
        process_data(stock_data())
      
      # Based on the selected model, perform the corresponding model fitting
      model_name <- input$model_name
      if (model_name == "Linear Regression") {
        df_model <- processed_data
        
        index <- floor(0.8 * nrow(df_model))
        
        # Split the data into train and test sets
        train_df <- df_model[1:index, ]
        test_df <- df_model[(index + 1):nrow(df_model), ]
        # Perform linear regression model fitting
        lm.model <- lm(ClosingPrice   ~ . - Date , data = train_df)
        
        output$model_prediction <- renderPlotly({
          req(test_df)  # Ensure test_df is available
          req(lm.model)  # Ensure lm.model is available
          
          # Check if fitting_done is TRUE before rendering the plot
          
          # Predict on the test data using the fitted linear regression model
          test_df$predicted <-
            predict(lm.model, newdata = test_df)
          
          
          
          # Create the plotly chart
          plot_ly(data = test_df, x = ~ Date) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
        
        
        
        
        
      } else if (model_name == "ARIMA") {
        date <- input$date_range[1]
        year <- year(date)
        month <- month(date, label = TRUE)
        
        processed_data <- process_data(stock_data())
        # convert to time series object
        ts_ko <-
          ts(
            processed_data$ClosingPrice,
            frequency = 365.25,
            start = c(year, month)
          )
        
        
        
        # Calculate the number of observations for the train set
        train_size <- floor(0.8 * length(ts_ko))
        
        # Perform the train-test split
        train <- head(ts_ko, train_size)
        test <- tail(ts_ko, length(ts_ko) - train_size)
        
        
        model.arima <-
          Arima(
            train,
            order = c(
              input$arima_order_1,
              input$arima_order_2,
              input$arima_order_3
            ),
            seasonal = list(
              order = c(
                input$arima_seasonal_order_1,
                input$arima_seasonal_order_2,
                input$arima_seasonal_order_3
              ),
              period = input$arima_seasonal_period
            ),
            lambda = input$arima_lambda
          )
        
        
        # predictions_test_lm <- predict(lm.model, newdata = test_df)
        
        output$model_prediction <- renderPlotly({
          req(test)  # Ensure test_df is available
          req(model.arima)  # Ensure lm.model is available
          
          # Convert test to a dataframe with dates
          test_df <-
            data.frame(Actual = coredata(test), Date = index(test))
          
          
          # Predict on the test data using the fitted ARIMA model
          test_forecast <- forecast(model.arima, h = nrow(test_df))
          test_df$Predicted <- as.numeric(test_forecast$mean)
          
          # Create the plotly chart
          plot_ly(data = test_df, x = ~ Date) %>%
            add_lines(
              y = ~ Actual,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ Predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
        
      } else if (model_name == "GARCH") {
        date <- input$date_range[1]
        year <- year(date)
        month <- month(date, label = TRUE)
        
        processed_data <- process_data(stock_data())
        # convert to time series object
        ts_ko <-
          ts(
            processed_data$ClosingPrice,
            frequency = 365.25,
            start = c(year, month)
          )
        
        # Calculate the number of observations for the train set
        train_size <- floor(0.8 * length(ts_ko))
        
        # Perform the train-test split
        train <- head(ts_ko, train_size)
        test <- tail(ts_ko, length(ts_ko) - train_size)
        
        
        model.spec <-
          ugarchspec(
            variance.model = list(
              model = "sGARCH",
              garchOrder = c(input$garch_order_1, input$garch_order_2)
            ),
            mean.model = list(
              armaOrder = c(input$garch_arma_order_1, input$garch_arma_order_2)
            ),
            distribution.model = "std"
          )
        
        
        # Fit the GARCH model to the data
        model.garch <-
          ugarchfit(
            spec = model.spec,
            data = train,
            solver = "hybrid",
            solver.control = list(maxit = 1000)
          )
        
        
        output$model_prediction <- renderPlotly({
          req(test)  # Ensure test_df is available
          req(model.garch)  # Ensure lm.model is available
          
          # Convert test to a dataframe with dates
          test_df <-
            data.frame(Actual = coredata(test), Date = index(test))
          
          # Generate forecasts for the test dataset
          test_forecast <-
            ugarchforecast(model.garch, n.ahead = length(test))
          
          
          
          
          # Predict on the test data using the fitted ARIMA model
          test_df$Predicted <- sigma(test_forecast)
          # Create the plotly chart
          plot_ly(data = test_df, x = ~ Date) %>%
            add_lines(
              y = ~ Actual,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ Predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
      } else if (model_name == "Random Forest") {
        processed_data <- process_data(stock_data())
        df_no_na <-  processed_data %>%
          select(-Date)
        
        
        index <- floor(0.8 * nrow(df_no_na))
        
        
        # Split the data into train and test sets
        train_df_na <- df_no_na[1:index, ]
        test_df_na <-
          processed_data[(index + 1):nrow(processed_data), ]
        
        
        
        
        X <- train_df_na %>%
          dplyr::select(sma_20 ,
                        sma_50 ,
                        rsi,
                        macd ,
                        returns ,
                        log_returns,
                        date_num)
        y <- train_df_na$ClosingPrice
        
        rf.model <-
          randomForest(ClosingPrice    ~ .,
                       data = train_df_na,
                       mtry = input$rf_mtry)
        
        
        
        output$model_prediction <- renderPlotly({
          req(test_df_na)  # Ensure test_df is available
          req(rf.model)  # Ensure lm.model is available
          
          # Check if fitting_done is TRUE before rendering the plot
          
          # Predict on the test data using the fitted linear regression model
          test_df_na$predicted <-
            predict(rf.model, newdata = test_df_na)
          
          
          
          # Create the plotly chart
          plot_ly(data = test_df_na, x = ~ Date) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
        
      } else if (model_name == "SVM") {
        processed_data <- process_data(stock_data())
        df_no_na <-  processed_data %>%
          select(-Date)
        
        
        index <- floor(0.8 * nrow(df_no_na))
        
        
        # Split the data into train and test sets
        train_df_na <- df_no_na[1:index, ]
        test_df_na <-
          processed_data[(index + 1):nrow(processed_data), ]
        
        
        
        
        X <- train_df_na %>%
          dplyr::select(sma_20 ,
                        sma_50 ,
                        rsi,
                        macd ,
                        returns ,
                        log_returns,
                        date_num)
        y <- train_df_na$ClosingPrice
        
        svm.model <- svm(
          ClosingPrice ~ .,
          data = train_df_na,
          cost = input$svm_c,
          sigma = input$svm_sigma
        )
        
        
        
        output$model_prediction <- renderPlotly({
          req(test_df_na)  # Ensure test_df is available
          req(svm.model)  # Ensure lm.model is available
          
          # Check if fitting_done is TRUE before rendering the plot
          
          # Predict on the test data using the fitted linear regression model
          test_df_na$predicted <-
            predict(svm.model, newdata = test_df_na)
          
          
          
          # Create the plotly chart
          plot_ly(data = test_df_na, x = ~ Date) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
      } else if (model_name == "Gradient Boosting") {
        processed_data <- process_data(stock_data())
        index <- floor(0.8 * nrow(processed_data))
        
        # Split the data into train and test sets
        train_df_na <- processed_data[1:index,]
        test_df_na <-
          processed_data[(index + 1):nrow(processed_data),]
        
        train_Dmatrix <- train_df_na %>%
          dplyr::select(date_num,
                        sma_20,
                        sma_50,
                        rsi,
                        macd,
                        returns,
                        log_returns) %>%
          as.matrix()
        
        test_Dmatrix <- test_df_na %>%
          dplyr::select(date_num,
                        sma_20,
                        sma_50,
                        rsi,
                        macd,
                        returns,
                        log_returns) %>%
          as.matrix()
        
        dtrain <-
          xgb.DMatrix(train_Dmatrix,
                      label = train_df_na$ClosingPrice,
                      nthread = 2)
        dtest <-
          xgb.DMatrix(test_Dmatrix,
                      label = test_df_na$ClosingPrice,
                      nthread = 2)
        
        # Set the model parameters
        param  <- list(
          eta = input$gb_eta,
          lambda = input$gb_lambda,
          alpha = input$gb_alpha
        )
        
        # Train the xgboost model
        watchlist <- list(train = dtrain, eval = dtest)
        
        xgb_model <- xgb.train(param, dtrain, nrounds = 2, watchlist)
        
        
        
        output$model_prediction <- renderPlotly({
          req(test_df_na)  # Ensure test_df_na is available
          req(xgb_model)  # Ensure xgb_model is available
          
          # Predict on the test data using the fitted xgboost model
          test_df_na$Predicted <-
            predict(xgb_model, dtest)
          
          # Create the plotly chart
          plot_ly(data = test_df_na, x = ~ date_num) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ Predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
        })
        
      } else if (model_name == "Recurrent Neural Networks") {
        processed_data <-
          process_data(stock_data())
        
        df_nn <- processed_data %>%
          dplyr::select(-Date)
        
        
        # Prepare the data_rnn
        data_rnn <- df_nn 
        
        # Split the data_rnn into training and testing sets
        train_size <- 0.8  # Percentage of data_rnn for training
        train_index <- round(train_size * nrow(data_rnn))
        train_data_rnn <- data_rnn[1:train_index,]
        test_data_rnn <-
          data_rnn[(train_index + 1):nrow(data_rnn),]
        
        test_data_rnn_plot <-
          processed_data[(train_index + 1):nrow(processed_data),]
        
        # Prepare the training and testing data_rnn
        lookback <- 5  # Number of previous time steps to consider
        # # Normalize the data_rnn
        train_data_rnn <- as.data.frame(scale(train_data_rnn))
        test_data_rnn <- as.data.frame(scale(test_data_rnn))
        # Initialize train_sequences_rnn and test_sequences_rnn
        train_sequences_rnn <-
          array(0, dim = c(
            nrow(train_data_rnn) - lookback + 1,
            lookback,
            ncol(train_data_rnn)
          ))
        test_sequences_rnn <-
          array(0, dim = c(
            nrow(test_data_rnn) - lookback + 1,
            lookback,
            ncol(test_data_rnn)
          ))
        
        # Populate train_sequences_rnn with subsets of train_data
        for (i in 1:(nrow(train_data_rnn) - lookback + 1)) {
          train_sequences_rnn[i, ,] <-
            as.matrix(train_data_rnn[i:(i + lookback - 1),])
        }
        
        # Populate test_sequences_rnn with subsets of test_data_rnn
        for (i in 1:(nrow(test_data_rnn) - lookback + 1)) {
          test_sequences_rnn[i, ,] <-
            as.matrix(test_data_rnn[i:(i + lookback - 1),])
        }
        
        
        # Define the rnn_model
        rnn_model <- keras_model_sequential()
        rnn_model %>%
          layer_simple_rnn (units = input$rnn_units,
                            input_shape = c(lookback, ncol(data_rnn))) %>%
          layer_dense(units = 1)
        
        # Compile the rnn_model
        rnn_model %>% compile(loss = 'mean_squared_error',
                              optimizer = 'adam')
        # Train the rnn_model
        epochs <- 100
        batch_size <- 32
        early_stopping <- callback_early_stopping(
          monitor = "loss",
          patience = 5,
          restore_best_weights = TRUE
        )
        
        rnn_model %>% fit(
          train_sequences_rnn,
          train_data_rnn$ClosingPrice[lookback:nrow(train_data_rnn)],
          epochs = epochs,
          batch_size = batch_size,
          verbose = 1,
          callbacks = list(early_stopping)
        )
        
        # Evaluate the rnn_model
        evaluation <- rnn_model %>% evaluate(test_sequences_rnn,
                                             test_data_rnn$ClosingPrice[lookback:nrow(test_data_rnn)],
                                             verbose = 0)
        
        print("Model Fitted")
        output$model_prediction <- renderPlotly({
          req(test_data_rnn)  # Ensure test_df is available
          req(rnn_model)  # Ensure lm.model is available
          
          
          # Make predictions
          test_predictions_rnn <-
            rnn_model %>% predict(test_sequences_rnn)
          
          # Print the evaluation results and predicted values
          print(head(test_predictions_rnn))
          
          # Convert test_predictions to a dataframe
          test_predictions_df_rnn <- data.frame(
            Date = tail(
              test_data_rnn$date_num,
              nrow(test_data_rnn) - lookback
            ),
            Actual = tail(
              test_data_rnn$ClosingPrice,
              nrow(test_data_rnn) - lookback
            ),
            Predicted = tail(
              test_predictions_rnn,
              nrow(test_data_rnn) - lookback
            )
          )
          
          # Scale the actual and predicted values back to their original scale
          actual_unscaled_test_rnn <-
            test_predictions_df_rnn$Actual * sd(df_nn$ClosingPrice) + mean(df_nn$ClosingPrice)
          predicted_unscaled_test_rnn <-
            test_predictions_df_rnn$Predicted * sd(df_nn$ClosingPrice) + mean(df_nn$ClosingPrice)
          
          # Create a new data frame with unscaled values
          unscaled_predictions_df_test_rnn <- data.frame(
            Date = tail(
              test_data_rnn_plot$Date,
              nrow(test_data_rnn) - lookback
            ),
            ClosingPrice = actual_unscaled_test_rnn,
            Predicted = predicted_unscaled_test_rnn
          )
          
          
          
          # Create the plotly chart
          plot_ly(data = unscaled_predictions_df_test_rnn, x = ~ Date) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ Predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
      } else if (model_name == "LSTM") {
        processed_data <-
          process_data(stock_data())
        
        df_nn <- processed_data %>%
          dplyr::select(-Date)
        
        
        # Prepare the data_lstm
        data_lstm <-
          df_nn  
        
        # Split the data_lstm into training and testing sets
        train_size <- 0.8  # Percentage of data_lstm for training
        train_index <- round(train_size * nrow(data_lstm))
        train_data_lstm <- data_lstm[1:train_index,]
        test_data_lstm <-
          data_lstm[(train_index + 1):nrow(data_lstm),]
        
        test_data_lstm_plot <-
          processed_data[(train_index + 1):nrow(processed_data),]
        
        # Prepare the training and testing data_lstm
        lookback <- 5  # Number of previous time steps to consider
        # # Normalize the data_lstm
        train_data_lstm <- as.data.frame(scale(train_data_lstm))
        test_data_lstm <- as.data.frame(scale(test_data_lstm))
        # Initialize train_sequences_lstm and test_sequences_lstm
        train_sequences_lstm <-
          array(0, dim = c(
            nrow(train_data_lstm) - lookback + 1,
            lookback,
            ncol(train_data_lstm)
          ))
        test_sequences_lstm <-
          array(0, dim = c(
            nrow(test_data_lstm) - lookback + 1,
            lookback,
            ncol(test_data_lstm)
          ))
        
        # Populate train_sequences_lstm with subsets of train_data
        for (i in 1:(nrow(train_data_lstm) - lookback + 1)) {
          train_sequences_lstm[i, ,] <-
            as.matrix(train_data_lstm[i:(i + lookback - 1),])
        }
        
        # Populate test_sequences_lstm with subsets of test_data_lstm
        for (i in 1:(nrow(test_data_lstm) - lookback + 1)) {
          test_sequences_lstm[i, ,] <-
            as.matrix(test_data_lstm[i:(i + lookback - 1),])
        }
        
        
        # Define the LSTM lstm_model
        lstm_model <- keras_model_sequential()
        lstm_model %>%
          layer_lstm(units = input$lstm_units,
                     input_shape = c(lookback, ncol(data_lstm))) %>%
          layer_dense(units = 1)
        
        # Compile the lstm_model
        lstm_model %>% compile(loss = 'mean_squared_error',
                               optimizer = 'adam')
        
        # Train the lstm_model
        epochs <- 100
        batch_size <- 32
        early_stopping <- callback_early_stopping(
          monitor = "loss",
          patience = 5,
          restore_best_weights = TRUE
        )
        
        lstm_model %>% fit(
          train_sequences_lstm,
          train_data_lstm$ClosingPrice[lookback:nrow(train_data_lstm)],
          epochs = epochs,
          batch_size = batch_size,
          verbose = 1,
          callbacks = list(early_stopping)
        )
        
        # Evaluate the lstm_model
        evaluation <- lstm_model %>% evaluate(test_sequences_lstm,
                                              test_data_lstm$ClosingPrice[lookback:nrow(test_data_lstm)],
                                              verbose = 0)
        
        print("Model Fitted")
        output$model_prediction <- renderPlotly({
          req(test_data_lstm)  # Ensure test_df is available
          req(lstm_model)  # Ensure lm.model is available
          
          
          # Make predictions
          test_predictions_lstm <-
            lstm_model %>% predict(test_sequences_lstm)
          
          # Print the evaluation results and predicted values
          print(head(test_predictions_lstm))
          
          # Convert test_predictions to a dataframe
          test_predictions_df_lstm <- data.frame(
            Date = tail(
              test_data_lstm$date_num,
              nrow(test_data_lstm) - lookback
            ),
            Actual = tail(
              test_data_lstm$ClosingPrice,
              nrow(test_data_lstm) - lookback
            ),
            Predicted = tail(
              test_predictions_lstm,
              nrow(test_data_lstm) - lookback
            )
          )
          
          # Scale the actual and predicted values back to their original scale
          actual_unscaled_test_lstm <-
            test_predictions_df_lstm$Actual * sd(df_nn$ClosingPrice) + mean(df_nn$ClosingPrice)
          predicted_unscaled_test_lstm <-
            test_predictions_df_lstm$Predicted * sd(df_nn$ClosingPrice) + mean(df_nn$ClosingPrice)
          
          # Create a new data frame with unscaled values
          unscaled_predictions_df_test_lstm <- data.frame(
            Date = tail(
              test_data_lstm_plot$Date,
              nrow(test_data_lstm) - lookback
            ),
            ClosingPrice = actual_unscaled_test_lstm,
            Predicted = predicted_unscaled_test_lstm
          )
          
          
          
          # Create the plotly chart
          plot_ly(data = unscaled_predictions_df_test_lstm, x = ~ Date) %>%
            add_lines(
              y = ~ ClosingPrice,
              name = "Actual",
              colors = c("Actual" = "blue"),
              width = 2
            ) %>%
            add_lines(
              y = ~ Predicted,
              name = "Predicted",
              colors = c("Predicted" = "red"),
              width = 2,
              dash = "dash"
            ) %>%
            layout(
              title = "Model Prediction",
              xaxis = list(
                title = "Date",
                tickfont = list(weight = "bold")
              ),
              yaxis = list(
                title = "Closing Price",
                tickfont = list(weight = "bold")
              ),
              showlegend = TRUE,
              legend = list(x = 0.1, y = 0.9)
            )
          
        })
        
      }
      
     
      # Enable the Fit Model button
      shinyjs::enable("fit_model_button")
      # Update "Model Fitting Done" text
      output$fit_model_status <- renderText("Model Fitting Done")
      # Set model_fitted to TRUE
      model_fitted(TRUE)
    } else {
      showNotification("Please download stock data first.", type = "warning")
      model_fitted(FALSE)
    }
  })
  
  # Disable Fit Model button when data is not downloaded or model is already fitted
  observe({
    shinyjs::disable("fit_model_button")  # Disable the button by default
    if (data_downloaded()) {
      shinyjs::enable("fit_model_button")  # Enable the button when conditions are met
    }
  })
  
  
  
})

shinyApp(ui, server)
